unit TestZBar1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls, ExtCtrls;

type

  { TTestZBarForm }

  TTestZBarForm = class(TForm)
    btnDecode: TButton;
    btnOpenFile: TButton;
    gbDecode: TGroupBox;
    Image1: TImage;
    mDecText: TMemo;
    procedure btnDecodeClick(Sender: TObject);
    procedure btnOpenFileClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    { private declarations }
    procedure LoadImage(const FileName: string);
  public
    { public declarations }
  end;

var
  TestZBarForm: TTestZBarForm;

implementation

uses
  zbar_qrdec;

{$R *.lfm}

{ TTestZBarForm }

procedure TTestZBarForm.FormShow(Sender: TObject);
begin
  if FileExists('test.jpg') then
    LoadImage('test.jpg');
end;

procedure TTestZBarForm.LoadImage(const FileName: string);
var bitmap: TBitmap;
begin
  Image1.Picture.LoadFromFile(FileName);
  bitmap:=TBitmap.Create;
  bitmap.PixelFormat := pf24bit;
  bitmap.Width := Image1.Picture.Bitmap.Width;
  bitmap.Height := Image1.Picture.Bitmap.Height;
  bitmap.Canvas.Draw(0, 0, Image1.Picture.Bitmap);
  Image1.Picture.Bitmap.Assign(bitmap);
  bitmap.Free;
end;

procedure TTestZBarForm.btnOpenFileClick(Sender: TObject);
var dlg: TOpenDialog;
begin
  dlg := TOpenDialog.Create(Self);
  dlg.Filter:='Image files (*.bmp,*.jpeg,*.jpg,*.gif,*.png)|*.BMP;*.JPEG;*.JPG;*.GIF;*.PNG';;
  if dlg.Execute then
    LoadImage(dlg.FileName);
  dlg.Free;
end;

procedure TTestZBarForm.btnDecodeClick(Sender: TObject);
var
  zimage: pzbar_image_t;
  zprocessor : pzbar_processor_t;
  zsymbol: pzbar_symbol_t;
  ztype: zbar_symbol_type_t;
  zpoints: array[0..3] of TPoint;
  grayimage: PByte;
  i,j: integer;

  procedure BitmapToGrayImage(dst: PByte);
  var bitmap: TBitmap; y,x: integer; src: PByte;
  begin
    bitmap:=TBitmap.Create;
    try
      // Use 24bit bitmap
      bitmap.PixelFormat := pf24bit;
      bitmap.Width := Image1.Picture.Bitmap.Width;
      bitmap.Height := Image1.Picture.Bitmap.Height;
      bitmap.Canvas.Draw(0, 0, Image1.Picture.Bitmap);

      for y := 0 to bitmap.Height - 1 do begin
        src := bitmap.ScanLine[y];
        for x := 0 to bitmap.Width - 1 do begin
          // 77(R) + 151(G) + 28(B) = 256
          dst^ := (src[2]*77 + src[1]*151 + src[0]*28) shr 8;
          Inc(src,3);
          Inc(dst);
        end;
      end;
    finally
      bitmap.Free;
    end;
  end;

begin
  mDecText.Lines.Clear;

  // "HAVE_zbar_qr_decode" must be defined (see zbar_qrdec.pas)
  zbar_set_verbosity(0);
  // processor
  zprocessor := zbar_processor_create(0);
  if not Assigned(zprocessor) then raise Exception.Create('zbar_processor_create failed!');
  zbar_processor_init(zprocessor, nil{video_device}, 0{enable_display});
  zbar_processor_set_visible(zprocessor, 0);
  // image
  zimage := zbar_image_create();
  if not Assigned(zimage) then raise Exception.Create('zbar_image_create failed!');
  zbar_image_set_format(zimage, 'Y800'{gray});
  zbar_image_set_size(zimage, Image1.Width, Image1.Height);
  grayimage:=GetMem(Image1.Width*Image1.Height);
  BitmapToGrayImage(grayimage);
  zbar_image_set_data(zimage, grayimage, Image1.Width*Image1.Height, nil);
  // decode
  zbar_process_image(zprocessor, zimage);
  zsymbol := zbar_image_first_symbol(zimage);
  i:=0;
  while Assigned(zsymbol) do begin
    ztype := zbar_symbol_get_type(zsymbol);
    if ztype = ZBAR_PARTIAL then continue;
    // bounding box
    for j:=0 to 3 do
      zpoints[j] := Point(zbar_symbol_get_loc_x(zsymbol,j), zbar_symbol_get_loc_y(zsymbol,j));
    // draw rectangle around found detected QR code
    with Image1 do begin
      Canvas.Pen.Color:=clRed;
      Canvas.Pen.Width:=1;
      Canvas.Pen.Style:=psSolid;
      Canvas.Polyline([zpoints[0], zpoints[1], zpoints[2], zpoints[3], zpoints[0]]);
    end;
    // decoded text
    mDecText.Lines.Append(Format('[%d]: %s', [i, StrPas(zbar_symbol_get_data(zsymbol))]));
    zsymbol := zbar_symbol_next(zsymbol);
    Inc(i);
  end;
  zbar_image_destroy(zimage);
  FreeMem(grayimage);
  zbar_processor_destroy(zprocessor);
end;

end.

