unit zbar_qrcode;

interface

const
  (*The number of bits of subpel precision to store image coordinates in.
    This helps when estimating positions in low-resolution images, which may have
     a module pitch only a pixel or two wide, making rounding errors matter a
     great deal.*)
  QR_FINDER_SUBPREC = 2;

type
  qr_point = array[0..1] of integer;
  pqr_point = ^qr_point;

  (*A line crossing a finder pattern.
    Whether the line is horizontal or vertical is determined by context.
    The offsts to various parts of the finder pattern are as follows:
      |*****|     |*****|*****|*****|     |*****|
      |*****|     |*****|*****|*****|     |*****|
         ^        ^                 ^        ^
         |        |                 |        |
         |        |                 |       pos[v]+len+eoffs
         |        |                pos[v]+len
         |       pos[v]
        pos[v]-boffs
    Here v is 0 for horizontal and 1 for vertical lines.*)
  qr_finder_line = record
    (*The location of the upper/left endpoint of the line.
      The left/upper edge of the center section is used, since other lines must
       cross in this region.*)
    pos: qr_point;
    (*The length of the center section.
      This extends to the right/bottom of the center section, since other lines
       must cross in this region.*)
    len: integer;
    (*The offset to the midpoint of the upper/left section (part of the outside
     ring), or 0 if we couldn't identify the edge of the beginning section.
    We use the midpoint instead of the edge because it can be located more
     reliably.*)
    boffs: integer;
    (*The offset to the midpoint of the end section (part of the outside ring),
       or 0 if we couldn't identify the edge of the end section.
      We use the midpoint instead of the edge because it can be located more
       reliably.*)
    eoffs: integer;
  end;
  pqr_finder_line = ^qr_finder_line;
  ppqr_finder_line = ^pqr_finder_line;

implementation

end.
