{$DEFINE HAVE_qrdectxt}

function text_is_ascii(_text: pbyte;  _len: integer): boolean;
var i: integer;
begin
  for i:=0 to Pred(_len) do
    if _text[i]>=$80 then Exit(False);
  Exit(True);
end;

function text_is_latin1(_text: pbyte;  _len: integer): boolean;
var i: integer;
begin
  for i:=0 to Pred(_len) do
    if (_text[i]>=$80)and(_text[i]<$A0) then Exit(False);
  Exit(True);
end;

function qr_code_data_list_extract_text(var _qrlist: qr_code_data_list;  iscn: pzbar_image_scanner_t;  img: pzbar_image_t): integer;
var
  qrdata, qrdataj: pqr_code_data;
  entry: pqr_code_data_entry;
  nqrdata: integer;
  mark: pbyte;
  sa: array [0..Pred(16)] of integer;
  sa_size: integer;
  sa_text: pbyte;
  sa_ntext, sa_ctext: integer;
  fnc1, fnc1_2ai: integer;
  has_kanji: boolean;
  eci, err: integer;
  i,j,k: integer;
  syms: pzbar_symbol_t;
  sym: ^pzbar_symbol_t;
  dir: qr_point;
  horiz: integer;
  sa_parity: unsigned;
  shift: integer;
  inleft, outleft, plen: size_t;
  _in, _out: pbyte;
  p: pbyte;
  c: char;
  ei,ej: integer;
  buf: array [0..Pred(16)] of char;
  cur_eci: unsigned;
  sa_sym: pzbar_symbol_t;
  xmin, xmax, ymin, ymax, next, u: integer;
begin
  qrdata := _qrlist.qrdata;
  nqrdata := _qrlist.nqrdata;
  mark := AllocMem(nqrdata*sizeof(mark^));
  Result := 0;
  (*This is a trivial conversion just to check validity without extra code.*)
  for i:=0 to Pred(nqrdata) do
    if mark[i] = 0 then begin
      syms:=nil;
      sym:=@syms;

     (*Step 0: Collect the other QR codes belonging to this S-A group.*)
     if qrdata[i].sa_size<>0 then begin
        sa_size:= qrdata[i].sa_size;
        sa_parity:= qrdata[i].sa_parity;
        for j:=0 to Pred(sa_size) do sa[j]:= -1;
        for j:=i to Pred(nqrdata) do if mark[j]=0 then begin
          (*TODO: We could also match version, ECC level, etc. if size and
                  parity alone are too ambiguous.*)
          if (qrdata[j].sa_size=sa_size) and (qrdata[j].sa_parity=sa_parity) and (sa[qrdata[j].sa_index]<0) then begin
            sa[qrdata[j].sa_index] := j;
            mark[j] := 1;
          end;
        end;
     end
     (*TODO: If the S-A group is complete, check the parity.*)
     else begin
       sa[0] := i;
       sa_size := 1;
     end;

     sa_ctext := 0;
     fnc1 := 0;
     fnc1_2ai := 0;
     has_kanji := false;
     (*Step 1: Detect FNC1 markers and estimate the required buffer size.*)
     for j:=0 to Pred(sa_size) do if sa[j]>=0 then begin
       qrdataj := qrdata+sa[j];
       {$IFDEF DEBUG_ZBAR}DebugLn('qr_code_data_list_extract_text: qrdata[%d] entries = %d',[j, qrdataj^.nentries]);{$ENDIF}
       for k:=0 to Pred(qrdataj^.nentries) do begin
         entry := qrdataj^.entries+k;
         shift := 0;
         case entry^.mode of
           (*FNC1 applies to the entire code and ignores subsequent markers.*)
           QR_MODE_FNC1_1ST:
            begin
              // #define MOD(mod) (1 << (mod))
              if fnc1=0 then
                fnc1 := 1 shl Ord(ZBAR_MOD_GS1);
            end;
           QR_MODE_FNC1_2ND:
            begin
              if fnc1=0 then begin
                fnc1 := 1 shl Ord(ZBAR_MOD_AIM);
                fnc1_2ai := entry^.payload.ai;
                Inc(sa_ctext, 2);
              end;
            end;
           else
            begin
              (*We assume at most 4 UTF-8 bytes per input byte.
                I believe this is true for all the encodings we actually use.*)
              case entry^.mode of
                QR_MODE_KANJI:
                  begin
                  has_kanji := True;
                  shift := 2; // *4
                  end;
                QR_MODE_BYTE:
                  shift := 2; // *4
              end;
              (*The remaining two modes are already valid UTF-8.*)
              if QR_MODE_HAS_DATA(entry^.mode) then
                Inc(sa_ctext, entry^.payload.data.len shl shift);
            end;
         end;
       end;
     end;

    (*Step 2: Convert the entries.*)
    {$IFDEF DEBUG_ZBAR}DebugLn('qr_code_data_list_extract_text: text size = %d',[sa_ctext]);{$ENDIF}
    sa_text := GetMem((sa_ctext+1)*sizeof(sa_text^));
    sa_ntext := 0;
    (*Add the encoded Application Indicator for FNC1 in the second position.*)
    if fnc1 = 1 shl Ord(ZBAR_MOD_AIM) then begin
      if fnc1_2ai<100 then begin
       (*The Application Indicator is a 2-digit number.*)
       sa_text[sa_ntext{++}] := Ord('0') + fnc1_2ai div 10;
       Inc(sa_ntext);
       sa_text[sa_ntext{++}] := Ord('0') + fnc1_2ai mod 10;
       Inc(sa_ntext);
      end
      else begin
       (*The Application Indicator is a single letter.
        We already checked that it lies in one of the ranges A...Z, a...z
        when we decoded it.*)
       sa_text[sa_ntext{++}] := fnc1_2ai-100;
       Inc(sa_ntext);
      end;
    end;

    eci := -1;
    err := 0;
    j := 0;
    while (j<sa_size) and (err=0) do begin
      sym^ := _zbar_image_scanner_alloc_sym(iscn,ZBAR_QRCODE_,0);
      (sym^)^.datalen := sa_ntext;

      if sa[j]<0 then begin
        (* generic placeholder for unfinished results *)
        (sym^)^.&type:=ZBAR_PARTIAL;
        (*Skip all contiguous missing segments.*)
        Inc(j);
        while (j<sa_size) and (sa[j]<0) do Inc(j);
        (*If there aren't any more, stop.*)
        if j>=sa_size then break;

        (* mark break in data *)
        sa_text[sa_ntext{++}]:= 0;
        Inc(sa_ntext);
        (sym^)^.datalen:=sa_ntext;

        (* advance to next symbol *)
        sym := @(sym^)^.next;
        sym^ := _zbar_image_scanner_alloc_sym(iscn,ZBAR_QRCODE_,0);
      end;

      qrdataj:= qrdata+sa[j];
      (* expose bounding box *)
      sym_add_point(sym^^, qrdataj^.bbox[0][0], qrdataj^.bbox[0][1]);
      sym_add_point(sym^^, qrdataj^.bbox[2][0], qrdataj^.bbox[2][1]);
      sym_add_point(sym^^, qrdataj^.bbox[3][0], qrdataj^.bbox[3][1]);
      sym_add_point(sym^^, qrdataj^.bbox[1][0], qrdataj^.bbox[1][1]);

      (* approx symbol "up" direction *)
      dir[0] := (qrdataj^.bbox[0][0] - qrdataj^.bbox[2][0] + qrdataj^.bbox[1][0] - qrdataj^.bbox[3][0]);
      dir[1] := (qrdataj^.bbox[2][1] - qrdataj^.bbox[0][1] + qrdataj^.bbox[3][1] - qrdataj^.bbox[1][1]);
      horiz := Ord(abs(dir[0]) > abs(dir[1]));
     (sym^)^.orient := zbar_orientation_t(horiz + 2*Ord(dir[1-horiz]<0));

     k:=0;
     while (k<qrdataj^.nentries) and (0=err) do begin
       entry:= qrdataj^.entries+k;
       case entry^.mode of
          QR_MODE_NUM:
            begin
              if sa_ctext-sa_ntext >= entry^.payload.data.len then begin
                Move(entry^.payload.data.buf^, sa_text[sa_ntext], entry^.payload.data.len*sizeof(sa_text^));
                Inc(sa_ntext, entry^.payload.data.len);
              end
              else
                err := 1;
            end;

          QR_MODE_ALNUM:
            begin
              _in := entry^.payload.data.buf;
              inleft := entry^.payload.data.len;
              (*FNC1 uses '%' as an escape character.*)
              if fnc1<>0 then while true do begin
                // TODO
                break;
              end
              else
                p:=nil;
              if (p<>nil) or (sa_ctext-sa_ntext<inleft) then
                err:= 1
              else begin
                Move(_in^, sa_text[sa_ntext], inleft*sizeof(sa_text^));
                Inc(sa_ntext, inleft);
              end;
            end;

          (*TODO: This will not handle a multi-byte sequence split between
                  multiple data blocks.
                  Does such a thing occur?
                  Is it allowed?
                  It requires copying buffers around to handle correctly.*)
          QR_MODE_BYTE,
          QR_MODE_KANJI:
            begin
              _in := entry^.payload.data.buf;
              inleft := entry^.payload.data.len;
              _out := sa_text+sa_ntext;
              outleft := sa_ctext-sa_ntext;
              (*If we have no specified encoding, attempt to auto-detect it.*)
              if eci<0 then begin
                if has_kanji then
                  //enc_list_mtf(enc_list,sjis_cd);
                else if (inleft>=3) and (_in[0]=$EF) and (_in[1]=$BB) and (_in[2]=$BF) then begin
                  (*Otherwise check for the UTF-8 BOM.
                    UTF-8 is rarely specified with ECI, and few decoders
                    currently support doing so, so this is the best way for
                    encoders to reliably indicate it.*)
                  Inc(_in, 3);
                  Dec(inleft, 3);
                  (*Actually try converting (to check validity).*)
                  //err := utf8_cd=(iconv_t)-1)or(iconv(utf8_cd,@in,@inleft,@out,@outleft)=(size_t)-1;
                  Assert(False, 'iconv()');
                  err := 0;
                  if err=0 then begin
                    sa_ntext := _out - sa_text;
                    //enc_list_mtf(enc_list,utf8_cd);
                    continue;
                  end;
                  _in := entry^.payload.data.buf;
                  inleft := entry^.payload.data.len;
                  _out := sa_text + sa_ntext;
                  outleft:= sa_ctext-sa_ntext;
                end
                (*If the text is 8-bit clean, prefer UTF-8 over SJIS, since
                  SJIS will corrupt the backslashes used for DoCoMo formats.*)
                else if text_is_ascii(_in,inleft) then begin
                  //enc_list_mtf(enc_list,utf8_cd);
                end;
                (*Try our list of encodings.*)
                // not implemented
                // iconv(): The iconv function returns the number of characters converted in a non-reversible way during this call
                //          it increments *inbuf and decrements *inbytesleft by the number of converted input bytes, it increments *outbuf and decrements *outbytesleft by the number of converted output bytes
                Move(_in^, _out^, inleft*sizeof(sa_text^));
                Inc(sa_ntext, inleft);
              end
              (*We were actually given a character set; use it.
                The spec says that in this case, data should be treated as if it
                 came from the given character set even when encoded in kanji
                 mode.*)
              else begin
                 Assert(False, 'iconv()');
                //err := eci_cd=(iconv_t)-1 || iconv(eci_cd,&in,&inleft,&out,&outleft)==(size_t)-1;
                //if (!err) sa_ntext=out-sa_text;
              end;
            end;
          QR_MODE_ECI:
            (*Check to see if a character set was specified.*)
            begin
             // not implemented
            end;
          else
            (*Silence stupid compiler warnings.*)
            begin
            end;
       end;{case}
       Inc(k);
     end;
     Inc(j);
     sym := @(sym^)^.next;
    end;

    if err=0 then begin
      sa_text[sa_ntext{++}]:= 0; Inc(sa_ntext);
      if sa_ctext+1>sa_ntext then
        sa_text:= ReallocMem(sa_text, sa_ntext*sizeof(sa_text^));

      if sa_size=1 then
        sa_sym:= syms
      else begin
        (* cheap out w/axis aligned bbox for now *)
        xmin:=img^.width;
        xmax:=-2;
        ymin:=img^.height;
        ymax:=-2;

        (* create "virtual" container symbol for composite result *)
        sa_sym := _zbar_image_scanner_alloc_sym(iscn,ZBAR_QRCODE_,0);
        sa_sym^.syms := _zbar_symbol_set_create();
        sa_sym^.syms^.head := syms;

        (* fixup data references *)
        while syms<>nil do begin
          _zbar_symbol_refcnt(syms,1);
          if syms^.&type=ZBAR_PARTIAL then
            sa_sym^.&type := ZBAR_PARTIAL
          else
            for j:=0 to Pred(syms^.npts) do begin
              u:=syms^.pts[j].x;
              if xmin>=u then xmin:= u-1;
              if xmax<=u then xmax:= u+1;
              u:= syms^.pts[j].y;
              if ymin>=u then ymin:= u-1;
              if ymax<=u then ymax:= u+1;
          end;
          syms^.data := sa_text + syms^.datalen;
          if Assigned(syms^.next) then
            next := syms^.next^.datalen
          else
            next := sa_ntext;
          assert(next>syms^.datalen);
          syms^.datalen := next - syms^.datalen - 1;

          syms := syms^.next; // for
        end;

        if xmax>=-1 then begin
          sym_add_point(sa_sym^, xmin, ymin);
          sym_add_point(sa_sym^, xmin, ymax);
          sym_add_point(sa_sym^, xmax, ymax);
          sym_add_point(sa_sym^, xmax, ymin);
        end;
      end;
      sa_sym^.data := sa_text;
      sa_sym^.data_alloc := sa_ntext;
      sa_sym^.datalen := sa_ntext-1;
      sa_sym^.modifiers := fnc1;
      _zbar_image_scanner_add_sym(iscn,sa_sym);
    end
    else begin
      _zbar_image_scanner_recycle_syms(iscn,syms);
      FreeMem(sa_text);
    end;
  end;
  FreeMem(mark);
end;
